/******************************************************************** 
   Copyright (C) 2000 Bassoukos Tassos <abas@aix.meng.auth.gr>
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later
   version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*********************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <gnome.h>
#include <glib.h>


#include "network.h"
#include "protocol.h"
#include "bookmarks.h"
#include "tasks.h"
#include "threads.h"
#include "server.h"
#include "login.h"
#include "smalltrans.h"
#include "news.h"
#include "userlist.h"
#include "pubchat.h"
#include "guiprefs.h"
#include "privs.h"
#include "filelist.h"
#include "files.h"
#include "filetransfer.h"

typedef struct {
  char *name;
  gpointer data;
  void (*create)(Connection *);
  void (*destroy)(Connection *,gpointer);
} Hook;

static Hook *hooks_find(Connection *c,char *name){
  GList *l;
  gboolean found=FALSE;
  
  g_mutex_lock(c->mutex);
  for(l=g_list_first(c->hooks);l!=NULL;l=g_list_next(l))
    if(strcmp(name,((Hook *)l->data)->name)==0){
      found=TRUE;
      break;
    }
  g_mutex_unlock(c->mutex);
  if(found==TRUE)
    return (Hook *)l->data;
  else 
    return NULL;
}

void hooks_create(Connection *c,char *name,
		void (*create)(Connection *),
		void (*destroy)(Connection *,gpointer)){
  Hook *h;

  if(hooks_find(c,name)!=NULL)
    return;
  h=(Hook *)malloc(sizeof(Hook));
  h->data=NULL;
  h->create=create;
  h->destroy=destroy;
  h->name=name;
  g_mutex_lock(c->mutex);
  c->hooks=g_list_append(c->hooks,h);
  g_mutex_unlock(c->mutex);
}

void hooks_set_data(Connection *c,char *name,gpointer data){
  Hook *h=hooks_find(c,name);
  if(h==NULL){
    printf(_("Attempted to set data for hook %s without init!\n"),name);
    return;
  }
  if(h->data!=NULL && data!=NULL)
    printf(_("Overwriting old data for hook %s..\n"),name);
  h->data=data;
}

gpointer hooks_get_data(Connection *c,char *name){
  Hook *h=hooks_find(c,name);
  if(h==NULL)
    return NULL;
  return h->data;
}

static void hooks_run_destroy(Connection *c){
  Hook *h;
  GList *l;
  for(l=g_list_first(c->hooks);l!=NULL;l=g_list_next(l)){
    h=(Hook*)l->data;
    if(h->data!=NULL && h->destroy!=NULL)
      h->destroy(c,h->data);
    h->data=NULL;
  }
}

static void hooks_run_create(Connection *c){
  Hook *h;
  GList *l;
  for(l=g_list_first(c->hooks);l!=NULL;l=g_list_next(l)){
    h=(Hook*)l->data;
    if(h->create!=NULL)
      h->create(c);
  }
}

/* ================================== */

GList *connections_list=NULL;

typedef struct {
  void (*start)(Connection *c,gpointer p);
  gpointer data;
} ProcEntry;

typedef struct {
  int task;
  HLTransaction *transaction;
  gpointer data;
  void (*start)(Connection *,HLTransaction *orig,HLTransaction *reply,gpointer p);
} PendingReply;

void server_transaction_start(Connection *c,
			      void (*start)(Connection *c,gpointer p),
			      gpointer data){
  ProcEntry *p=(ProcEntry *)malloc(sizeof(ProcEntry));
  p->start=start;
  p->data=data;
  g_mutex_lock(c->mutex);
  c->proclist=g_list_append(c->proclist,p);
  g_cond_signal(c->signal);
  g_mutex_unlock(c->mutex);
}

void server_transaction_reply_set(Connection *c,
				  HLTransaction *t,
				  void (*start)
				  (Connection *,HLTransaction *,HLTransaction *,gpointer),
				  gpointer data){
  PendingReply *p=(PendingReply *)malloc(sizeof(PendingReply));
  p->start=start;
  p->data=data;
  p->transaction=t;
  p->task=t->w.task;
  g_mutex_lock(c->mutex);
  c->replies_pending=g_list_append(c->replies_pending,p);
  g_cond_signal(c->signal);
  g_mutex_unlock(c->mutex);
}

void server_reply_default(Connection *c,HLTransaction *t,HLTransaction *reply,gpointer data){
  if(t->task)
    task_remove_when_idle(t->task);
  transaction_destroy(t);
  transaction_read_objects(c,reply);
  if(reply->w.error!=0){
    char buf[256];
    HLObject *o=transaction_find_object(reply,HLO_ERRORMSG);
    buf[0]=0;
    if(data!=NULL)
      strcpy(buf,(char *)data);
    if(o!=NULL){
      strcat(buf,"\n");
      strcat(buf,o->data.string);
    } else 
      strcat(buf,_("\nunspecified error"));
    show_error((c!=NULL && c->gui!=NULL && c->gui->main_window!=NULL)?c:NULL,
	       NULL,buf);
  }
  if(data!=NULL)
    free(data);
  transaction_destroy(reply);
}

void forall_servers(void (*func)(Connection *,gpointer),gpointer data){
  GList *l;
  for(l=g_list_first(connections_list);l!=NULL;l=g_list_next(l))
    func((Connection *)l->data,data);
}

static gboolean server_thread_ended(gpointer p){
  Connection *c=(Connection *)p;

  /* FIXME: check for reconnect ??? */
  /* FIXME: remove_lists */
  connections_list=g_list_remove(connections_list,c);
  if(c->infd!=NULL)
    fclose(c->infd);
  if(c->outfd!=NULL)
    fclose(c->outfd);
  hooks_run_destroy(c);
  if(c->gui!=NULL)
    gui_destroy(c);
  if(c->task!=NULL)
    task_remove(c->task);
  if(c->mutex)
    g_mutex_free(c->mutex);
  if(c->signal)
    g_cond_free(c->signal);
  free_bookmark(c->bookmark);
  free(c);
  return FALSE;
}

static void server_thread_end(gpointer p){
  Connection *c=(Connection *)p;
  gboolean start=FALSE;
  g_mutex_lock(c->mutex);
  if((--c->end_count)==0)
    start=TRUE;
  g_mutex_unlock(c->mutex);
  if(start==TRUE)
    gtk_idle_add(server_thread_ended,p);
}

static void server_thread_end_writer(gpointer p){
  Connection *c=(Connection *)p;
  g_mutex_unlock(c->mutex);
}

static gboolean server_do_stop(gpointer data){
  Connection *c=(Connection *)data;
  killThread(&c->thread_in);
  killThread(&c->thread_out);
  return FALSE;
}

static void server_user_cancel(Task *t){
  Connection *c=(Connection *)t->user_data;
  c->task=NULL;
  gtk_idle_add(server_do_stop,c);
}

void server_cancel(Connection *c){
  gtk_idle_add(server_do_stop,c);
}

static struct ServerMessage {
  int type;
  void (*handle)(Connection *c,HLTransaction *t,gpointer data);
  gpointer data;
  char *message;
} server_messages[]={
  {HLSI_AGREEMENT,handle_server_agreement,NULL,N_("agreement")},
  {HLSI_NEWPOST,handle_server_new_post,NULL,N_("news port")},
  {HLSI_USERCHANGE,handle_user_change,NULL,NULL},
  {HLSI_USERLEAVE,handle_user_leave,NULL,NULL},
  {HLSI_BROADCAST,handle_server_message,NULL,NULL},
  {HLSI_RELAYCHAT,pubchat_receive_mesasage,NULL,NULL},
  {HLSI_SELFUSER,handle_privs_sent,NULL,NULL},
  {HLSI_SERVERQUPDATE,handle_ftq_update,NULL,NULL},
  {0,NULL,NULL}
};

typedef struct {
  char *name;
  Connection *c;
  Task *t;
  GMutex *m;
  GCond *cond;
} TaskRequest;

static gboolean server_got_task(gpointer data){
  TaskRequest *tq=data;
  g_mutex_lock(tq->m);
  tq->t=task_new(tq->c->bookmark->host);
  task_add(tq->t);
  strcpy(tq->t->text,tq->name);
  task_update(tq->t,TRUE,0.1);
  g_cond_signal(tq->cond);
  g_mutex_unlock(tq->m);
  return FALSE;
}

static Task *server_get_task(Connection *c,struct ServerMessage *sm){
  volatile TaskRequest tq={NULL,NULL,NULL};
  char buf[2048];

  tq.c=c;
  tq.name=buf;
  sprintf(buf,"Retrieving %s",sm->message);
  tq.m=g_mutex_new();
  tq.cond=g_cond_new();
  g_mutex_lock(tq.m);
  gtk_idle_add(server_got_task,(gpointer)&tq);
  g_cond_wait(tq.cond,tq.m);
  g_mutex_unlock(tq.m);
  g_cond_free(tq.cond);
  g_mutex_free(tq.m);
  return tq.t;
}

static void server_handle_remote_request(Connection *c,HLTransaction *t){
  struct ServerMessage *sm=server_messages;
  int i;

  for(;sm->handle!=NULL;sm++)
    if(sm->type==t->w.type){
      Task *task=NULL;
      if(sm->message!=NULL){
	task=server_get_task(c,sm);
	t->task=task;
      }
      sm->handle(c,t,sm->data);
      if(task!=NULL)
	task_remove_when_idle(task);
      return;
    } 
  printf(_("Unknown Server-initiated transaction: info=%d type=%d task=%d\n"),
	 t->w.info,t->w.type,t->w.task);
  transaction_read_objects(c,t);
  printf(_(" object_count=%d;\n"),t->w.object_count);
  for(i=0;i<t->w.object_count;i++)
    printf(_(" %d.type=%d;\n"),i,t->objects[i]->type);
  transaction_destroy(t);
}

void server_set_default_reply(Connection *c,HLTransaction *t){
  server_transaction_reply_set(c,t,server_reply_default,NULL);
}

static void server_handle_reply(Connection *c,HLTransaction *t){
  PendingReply *pr=NULL;
  GList *l;
  gboolean found=FALSE;

  g_mutex_lock(c->mutex);
  l=g_list_first(c->replies_pending);
  do {
    if(l==NULL) break;
    
    if(l->data!=NULL && ((PendingReply *)l->data)->task == t->w.task){
      found=TRUE;
      pr=(PendingReply *)l->data;
      break;
    }
    l=g_list_next(l);
  } while(l!=NULL && l!=g_list_first(c->replies_pending));
  if(found==TRUE){
    c->replies_pending=g_list_remove_link(c->replies_pending,l);
    g_list_free_1(l);
  }
  g_mutex_unlock(c->mutex);
  if(found==FALSE){
    char buf[256];
    transaction_read_objects(c,t);
    sprintf(buf,_("Unexpected reply to task %d"),t->w.task);
    show_message_error(c,buf);
    transaction_destroy(t);
    return;
  }
  pr->start(c,pr->transaction,t,pr->data);
  free(pr);
}

static void server_do_read(gpointer p){
  Connection *c=(Connection *)p;
  HLTransaction *t;

  while(1){
    t=transaction_read_header(c);
    if(t==NULL){
      printf(_("connection closed, aborting ...\n"));
      server_transaction_start(c,NULL,NULL);
      return;
    }
    if(t->w.type==0){
      server_handle_reply(c,t);
    } else {
      server_handle_remote_request(c,t);
    }
  }
}


static void login_handle_reply(Connection *c,HLTransaction *t,
			       HLTransaction *r,gpointer p){
  Task *task;
  HLObject *o;

  if((task=c->task)!=NULL){
    c->task=NULL;
    task->user_data=NULL;
    task->user_cancel=NULL;
    task_remove_when_idle(task);
  }
  t->task=NULL;
  r->task=NULL;
  if(r->w.error){
    server_reply_default(c,t,r,strdup(_("Login Error:")));
    server_cancel(c);
    return;
  }
  transaction_read_objects(c,r);
  if((o=transaction_find_object(r,HLO_SERVERNAME))!=NULL){
    char *s=c->bookmark->name;
    c->bookmark->name=strdup(o->data.string);
    if(s!=NULL)
      free(s);
  }
  if((o=transaction_find_object(r,HLO_SERVERVERSION))!=NULL){
    c->version=o->data.number;
  }
  transaction_destroy(t);
  transaction_destroy(r);
  gtk_idle_add(login_gui_create,(gpointer)c);
}

static void start_login(Connection *c){
  HLTransaction *t;
  
  t=transaction_new(HLCT_LOGIN,(gpointer)c,TRUE);
  transaction_add_object(t,create_number(HLO_ICON,user_icon_number));
  if(c->bookmark->login!=NULL &&  strlen(c->bookmark->login)>0)
    transaction_add_object(t,create_string(HLO_LOGIN,c->bookmark->login));
  if(c->bookmark->passwd!=NULL && strlen(c->bookmark->passwd)>0)
    transaction_add_object(t,create_string(HLO_PASSWD,c->bookmark->passwd));
  transaction_add_object(t,create_string(HLO_NICK,nickname));
  transaction_add_object(t,create_number(HLO_SERVERVERSION,151));
  transaction_set_task(t,c->task,0.75,1.0);
  server_transaction_reply_set(c,t,login_handle_reply,(gpointer)c);
  transaction_send(t,c);
  ++c->end_count;
  newThread(&c->thread_in,server_do_read,server_thread_end,(gpointer)c);
}

static void server_do_write(gpointer p){
  Connection *c=(Connection *)p;
  NetAddr na;
  char buf[256];
  GList *l;
  ProcEntry *pe;

  if(isIP(&na,c->bookmark->host)==FALSE){
    if(resolveHost(&na,c->bookmark->host)==FALSE){
      sprintf(buf,_("No IP for name %s:"),c->bookmark->host);
      show_error_errno(c,NULL,buf);
      return;
    }
  }
  if(c->task!=NULL){
    sprintf(c->task->text,_("Connecting to %s..."),inet_ntoa(na));
    task_update(c->task,TRUE,0.25);
  } else 
    pthread_testcancel();
  c->infd=getSocketTo(&na,c->bookmark->port);
  if(c->infd==NULL){
    sprintf(buf,_("Could not connect to %s:%d :"),
	    c->bookmark->host,c->bookmark->port);
    show_error_errno(c,NULL,buf);
    return;
  }
  c->outfd=fdopen(dup(fileno(c->infd)),"w");

  if(c->task!=NULL){
    sprintf(c->task->text,_("Handshaking..."));
    task_update(c->task,TRUE,0.50);
  } else 
    pthread_testcancel();
  fwrite("TRTPHOTL\00\01\00\02",12,1,c->outfd);
  fflush(c->outfd);
  if(fread(buf,8,1,c->infd)!=1 ||
     memcmp(buf,"TRTP\0\0\0\0",8)!=0){
    sprintf(buf,_("%s is not a hotline server!"),c->bookmark->host);
    show_error(c,NULL,buf);
    return;
  }
  if(c->task!=NULL){
    sprintf(c->task->text,_("Logging in ..."));
    task_update(c->task,TRUE,0.75);
  } else 
   pthread_testcancel();

  start_login(c);
 
  while(1){
    g_mutex_lock(c->mutex);
    if((l=g_list_first(c->proclist))!=NULL){
      c->proclist=g_list_remove_link(c->proclist,l);
      g_mutex_unlock(c->mutex);
      pe=(ProcEntry *)l->data;
      g_list_free_1(l);
      if(pe->start!=NULL){
	pe->start(c,pe->data);
	free(pe);
      } else {
	free(pe);
	break;
      }
    } else {
      pthread_cleanup_push(server_thread_end_writer,c);
      g_cond_wait(c->signal,c->mutex);
      pthread_cleanup_pop(1);
    }
  }
}

void server_connect_to(Bookmark *b){
  Connection *c=(Connection *)malloc(sizeof(Connection));
  c->bookmark=dup_bookmark(b);
  c->infd=NULL;
  c->outfd=NULL;
  c->gui=NULL;
  c->version=0;
  c->signal=g_cond_new();
  c->mutex=g_mutex_new();
  c->can_read=FALSE;
  c->proclist=NULL;
  c->replies_pending=NULL;
  c->hooks=NULL;
  c->task=task_new(b->host);
  c->task->user_data=(gpointer)c;
  c->task->user_cancel=server_user_cancel;
  sprintf(c->task->text,_("Resolving %s ..."),b->host);
  task_update(c->task,TRUE,0.0);
  task_add(c->task);
  connections_list=g_list_append(connections_list,c);
  c->end_count=1;
  newThread(&c->thread_out,server_do_write,server_thread_end,(gpointer)c);
}
